#include "tilda.h"
#include "tilda-window.h"
#include "tilda-window-dbus-glue.h"

static gint
tilda_window_find_next_free_terminal_number (TildaWindow *tw)
{
	gint i, j;
	gboolean found;

	for (i=0; i<INT_MAX; ++i)
	{
		found = FALSE;

		for (j=0; j<tw->terms->len; ++j)
		{
			TildaTerminal *tt = g_ptr_array_index (tw->terms, j);

			if (tt->number == i)
			{
				found = TRUE;
				break;
			}
		}

		if (!found)
			return i;
	}

	return 0;
}

static gboolean
tilda_window_add_term (TildaWindow *tw)
{
	gint number;
	TildaTerminal *tt;

	number = tilda_window_find_next_free_terminal_number (tw);
	tt = g_object_new (TILDA_TYPE_TERMINAL,
					   "number", number,
					   "window-number", tw->number,
					   "parent-window", tw,
					   NULL);
	g_ptr_array_add (tw->terms, tt);

	GtkWidget *label = gtk_label_new ("Tilda");
	gint index = gtk_notebook_prepend_page (GTK_NOTEBOOK(tw->notebook), tt->hbox, label);
	gtk_notebook_set_tab_label_packing (GTK_NOTEBOOK(tw->notebook), tt->hbox, TRUE, TRUE, GTK_PACK_END);
	//gtk_notebook_set_current_page (GTK_NOTEBOOK(tw->notebook), index);

	if (gtk_notebook_get_n_pages (GTK_NOTEBOOK(tw->notebook)) > 1)
		gtk_notebook_set_show_tabs (GTK_NOTEBOOK(tw->notebook), TRUE);

	return TRUE;
}

/**
 * Remove the TildaTerminal with the given number from the given
 * TildaWindow.
 *
 * Return: TRUE on success, FALSE otherwise.
 */
gboolean
tilda_window_remove_term (TildaWindow *tw, gint terminal_number)
{
	gint i;

	for (i=0; i<tw->terms->len; ++i)
	{
		TildaTerminal *tt = g_ptr_array_index (tw->terms, i);

		if (tt->number == terminal_number)
		{
			gint notebook_index = gtk_notebook_page_num (GTK_NOTEBOOK(tw->notebook), tt->hbox);

			/* Make sure the index was valid */
			if (notebook_index == -1)
			{
				g_printerr ("DEBUG ERROR: Bad Notebook Tab\n");
				return FALSE;
			}

			/* Actually remove the terminal */
			gtk_notebook_remove_page (GTK_NOTEBOOK (tw->notebook), notebook_index);

			/* We should hide the tabs if there is only one tab left */
			if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook)) == 1)
				gtk_notebook_set_show_tabs (GTK_NOTEBOOK (tw->notebook), FALSE);

#if 0
			// FIXME FIXME FIXME: need to actually do the stuff below
			/* With no pages left, it's time to leave the program */
			if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (tw->notebook)) < 1)
				gtk_main_quit ();
#endif

			/* Remove the term from our lists, then free it */
			g_ptr_array_remove_fast (tw->terms, tt);
			g_object_unref (G_OBJECT(tt));

			/* Leave the loop, we're done */
			break;
		}
	}

	return TRUE;
}

/**
 * This sets up the given TildaWindow for the capability of real
 * transparency, if the X server is capable of it. */
static void
tilda_window_setup_real_transparency (TildaWindow *self)
{
	GdkScreen *screen;
	GdkColormap *colormap;

	screen = gtk_widget_get_screen (GTK_WIDGET(self->window));
	colormap = gdk_screen_get_rgba_colormap (screen);

	/* If possible, set the RGBA colormap so VTE can use real alpha
	 * channels for transparency. */
	if (colormap != NULL && gdk_screen_is_composited (screen))
	{
		gtk_widget_set_colormap (GTK_WIDGET(self->window), colormap);
		self->have_real_transparency = TRUE;
		return;
	}

	self->have_real_transparency = FALSE;
}

static void
tilda_window_dbus_register_object (TildaWindow *tw)
{
	gchar *object_path;

	// Register this object with DBus
	object_path = g_strdup_printf ("/net/sourceforge/Tilda/Window%d", tw->number);
	dbus_g_connection_register_g_object (dbus_connection, object_path, G_OBJECT(tw));
	g_free (object_path);
}

/*******************************************************************************
 * ALL GOBJECT STUFF BELOW PLEASE
 ******************************************************************************/

static GObjectClass *parent_class = NULL;

enum tilda_window_properties {
	TILDA_WINDOW_NUMBER = 1,

	TILDA_WINDOW_HAVE_REAL_TRANSPARENCY,
};

static void
tilda_window_instance_init (GTypeInstance *instance,
							gpointer       g_class)
{
	TildaWindow *self = (TildaWindow *) instance;
	self->dispose_has_run = FALSE;

	/* Initialize all properties */
	self->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	self->notebook = gtk_notebook_new ();
	self->terms = g_ptr_array_new ();

	/* Somewhat of a "poison" value, incase we don't set this */
	self->number = 0xdeadbeef;
}

static void
tilda_window_set_property (GObject      *object,
						   guint         property_id,
						   const GValue *value,
						   GParamSpec   *pspec)
{
	TildaWindow *self = (TildaWindow *) object;

	switch (property_id) {

		case TILDA_WINDOW_NUMBER:
			self->number = g_value_get_int (value);
			g_print ("window number: %d\n", self->number);
			break;

		case TILDA_WINDOW_HAVE_REAL_TRANSPARENCY:
			self->have_real_transparency = g_value_get_boolean (value);
			g_print ("window have real transp: %d\n", self->have_real_transparency);
			break;

		default:
			/* We don't have this property */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
	}
}

static void
tilda_window_get_property (GObject    *object,
						   guint       property_id,
						   GValue     *value,
						   GParamSpec *pspec)
{
	TildaWindow *self = (TildaWindow *) object;

	switch (property_id) {

		case TILDA_WINDOW_NUMBER:
			g_value_set_int (value, self->number);
			break;

		case TILDA_WINDOW_HAVE_REAL_TRANSPARENCY:
			g_value_set_boolean (value, self->have_real_transparency);
			break;

		default:
			/* We don't have this property */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
	}
}

static GObject *
tilda_window_constructor (GType                  type,
						  guint                  n_construct_properties,
						  GObjectConstructParam *construct_properties)
{
	GObject *obj;
	TildaWindow *self;

	/* Invoke parent constructor */
	TildaWindowClass *klass;
	klass = TILDA_WINDOW_CLASS (g_type_class_peek (TILDA_TYPE_WINDOW));
	obj = parent_class->constructor (type,
									 n_construct_properties,
									 construct_properties);

	/* Do other stuff here. The object is ready to go now, and all
	 * ctor properties have been set.
	 *
	 * TODO: This is the place to do DBus-init */
	self = TILDA_WINDOW(obj);

	/* Register this object with DBus */
	tilda_window_dbus_register_object (self);

	/* Try to set up real transparency */
	tilda_window_setup_real_transparency (self);

	gtk_container_add (GTK_CONTAINER(self->window), self->notebook);
	gtk_widget_show (self->notebook);

	tilda_window_add_term (self);
	tilda_window_add_term (self);
	gtk_widget_show_all (self->window);

	return obj;
}

static void
my_unref (gpointer data, gpointer user_data)
{
	g_object_unref (G_OBJECT(data));
}

static void
tilda_window_dispose (GObject *obj)
{
	TildaWindow *self = (TildaWindow *) obj;

	/* We don't want to run dispose twice, so just return immediately */
	if (self->dispose_has_run)
		return;

	/*
	 * In dispose, you are supposed to free all types referenced from this
	 * object which might themselves hold a reference to self. Generally,
	 * the most simple solution is to unref all members on which you own a
	 * reference.
	 *
	 * NOTE: See the following for how to deal with GtkObject-derived things:
	 * http://library.gnome.org/devel/gtk/unstable/GtkObject.html
	 */
	g_ptr_array_foreach (self->terms, my_unref, NULL);
	gtk_widget_destroy (self->window);

	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

static void
tilda_window_finalize (GObject *obj)
{
	TildaWindow *self = (TildaWindow *) obj;

	/*
	 * Here, complete the object's destruction.
	 * You might not need to do much...
	 */
	// TODO: g_free() any primitives here
	g_ptr_array_free (self->terms, TRUE);


	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static void
tilda_window_class_init (gpointer g_class,
						 gpointer g_class_data)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	TildaWindowClass *klass = TILDA_WINDOW_CLASS (g_class);
	GParamSpec *pspec;

	/* Hook our functions to this type */
	gobject_class->set_property = tilda_window_set_property;
	gobject_class->get_property = tilda_window_get_property;
	gobject_class->dispose = tilda_window_dispose;
	gobject_class->finalize = tilda_window_finalize;
	gobject_class->constructor = tilda_window_constructor;

	parent_class = g_type_class_peek_parent (klass);

	/* Install all of the properties */
	pspec = g_param_spec_int ("number",
							  "Window number",
							  "Set window's number",
							  0,		// min value
							  INT_MAX,	// max value
							  0,		// def value
							  G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_NUMBER,
									 pspec);

	pspec = g_param_spec_boolean ("have-real-transparency",
								  NULL, NULL, FALSE, G_PARAM_READABLE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_HAVE_REAL_TRANSPARENCY,
									 pspec);

	/* TODO: more properties */

	/* Hook the TildaWindow type into DBus */
	dbus_g_object_type_install_info (tilda_window_get_type(), &dbus_glib_tilda_window_object_info);
}

GType
tilda_window_get_type (void)
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo info = {
			sizeof (TildaWindowClass),
			NULL,	/* base_init */
			NULL,	/* base_finalize */
			tilda_window_class_init,	/* class_init */
			NULL,	/* class_finalize */
			NULL,	/* class_data */
			sizeof (TildaWindow),
			0,		/* n_preallocs */
			tilda_window_instance_init,	/* instance_init */
		};

		type = g_type_register_static (G_TYPE_OBJECT,
									   "TildaWindowType",
									   &info,
									   0);
	}

	return type;
}

#if 0

int main (int argc, char *argv[])
{
	GObject *tw;
	gint test_number = INT_MIN;

	/* Initialize the GObject type system */
	g_type_init ();
	gtk_init (&argc, &argv);

	tw = g_object_new (TILDA_TYPE_WINDOW, "number", 10, NULL);
	g_object_get (G_OBJECT (tw), "number", &test_number, NULL);
	g_assert (test_number == 10);

	g_object_unref (G_OBJECT (tw));

	tw = g_object_new (TILDA_TYPE_WINDOW, "number", 22, NULL);
	g_object_get (G_OBJECT (tw), "number", &test_number, NULL);
	g_assert (test_number == 22);

	gtk_main ();

	g_object_unref (G_OBJECT (tw));

	return 0;
}

#endif

/* vim: set ts=4 sts=4 sw=4 noet tw=112: */
